/*-
 * Copyright (c) 1993, 1994, 1995, 1996 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI ppp_ipcp.c,v 2.14 1996/12/31 23:59:30 prb Exp
 */

/*
 * This code is partially derived from CMU PPP.
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * PPP IPv6 Control Protocol
 */
#include <sys/types.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/slcompress.h>
#include <netinet6/in6_var.h>

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pbuf.h"
#include "ppp.h"
#include "ppp_proto.h"
#include "ppp_var.h"

static void   ppp_ipv6cp_setaddr(ppp_t *, struct sockaddr_in6 *, int);

void
ppp_ipv6cp_init(ppp)
	ppp_t *ppp;
{
	ppp->ppp_ncps++;
	ppp_ctl(ppp, PPP_SLCINIT);
	ppp_cp_up(ppp, FSM_IPV6CP);
}

int
ppp_ipv6cp_isup(ppp)
	ppp_t *ppp;
{
	return (ppp->ppp_ipv6cp.fsm_state == CP_OPENED ? 1 : 0);
}

int
ppp_ipv6cp_req(ppp)
	ppp_t *ppp;
{
	return (ppp_lcp_isup(ppp) && ppp_auth_isup(ppp));
}

/*
 * Timeout
 */
void
ppp_ipv6cp_timeout(ppp)
	void *ppp;
{
	ppp_cp_timeout((ppp_t *)ppp, FSM_IPV6CP);
}

/*
 * This level finished
 */
void
ppp_ipv6cp_tlf(ppp)
	ppp_t *ppp;
{
	/* If there is no more active NCPs drop the connection */
	if (--(ppp->ppp_ncps) == 0)
		ppp_down(ppp);
}

/*
 * IPCP reached CP_OPENED -- process the deferred queue
 */
void
ppp_ipv6cp_tlu(ppp)
	ppp_t *ppp;
{
	ppp_setparam(ppp, -1);		/* something might have changed */
	ppp_up(ppp);
	ppp_ctl(ppp, PPP_PROCDQ);
}

/*
 * Compose an IPCP Config-Request packet
 */
void
ppp_ipv6cp_creq(ppp)
	ppp_t *ppp;
{
	pbuf_t *m;
	pppopt_t opt;
	struct sockaddr_in6 a;
	u_long token;

	dprintf(D_PKT, "BUILD IPV6CP_CREQ:\n");
	m = 0;

#if 0
	/* IP-Compression-Protocol */
	if (!(ppp->ppp_ipv6cp.fsm_rej & (1<<IPV6CP_CPROT)) &&
	    (ppp->ppp_flags & PPP_TCPC)) {
		dprintf(D_DT, "	{CPROT: %x}\n", PPP_VJC);
		opt.po_type = IPV6CP_CPROT;
		opt.po_len = 6;
		opt.po_data.s = htons(PPP_VJC);
		opt.po_data.c[2] = MAX_STATES - 1;
		opt.po_data.c[3] = 0;
		m = ppp_addopt(m, &opt);
	}
#endif

	m = ppp_cp_prepend(m, PPP_IPV6CP, CP_CONF_REQ, (++ppp->ppp_ipv6cp.fsm_id), 0);
	sendpacket(ppp, m);
}

/*
 * Process the rejected option
 */
void
ppp_ipv6cp_optrej(ppp, opt)
	ppp_t *ppp;
	pppopt_t *opt;
{
	switch (opt->po_type) {
#if 0
	case IPV6CP_CPROT:
		dprintf(D_DT, "	{CPROT}\n");
		/*
		 * indicate that peer can't send compressed
		 */
		ppp->ppp_flags &= ~PPP_TCPC;
		break;
#endif

	default:
		dprintf(D_DT, "	{%d: len=%d}\n", opt->po_type, opt->po_len);
	}
}

/*
 * Process the NAK-ed option
 */
void
ppp_ipv6cp_optnak(ppp, opt)
	ppp_t *ppp;
	pppopt_t *opt;
{
	struct sockaddr_in6 a;

	switch (opt->po_type) {
#if 0
	case IPV6CP_CPROT:
		dprintf(D_DT, "	{CPROT: len=%d proto=0x%x}\n",
			opt->po_len, ntohs(opt->po_data.s));
		switch (ntohs(opt->po_data.s)) {
		case PPP_VJC:
			/*
			 * indicate that peer can't send compressed
			 */
			dprintf(D_DT, "	[Disable VJC]\n");
			ppp->ppp_flags &= ~PPP_TCPC;
			break;
		default:
			dprintf(D_DT, "	[Disable UNKNOWN]\n");
			break;
		}
		break;
#endif

	default:
		dprintf(D_DT, "	{%d: len=%d}\n", opt->po_type, opt->po_len);
	}
}

/*
 * Process the received option
 */
int
ppp_ipv6cp_opt(ppp, opt)
	ppp_t *ppp;
	pppopt_t *opt;
{
	struct sockaddr_in6 a;
	struct sockaddr_in6 b;
	int res;

	switch (opt->po_type) {
	default:
		dprintf(D_DT, "	{UNKNOWN: type=%d len=%d}\n",
		    opt->po_type, opt->po_len);
		return (OPT_REJ);

	case IPV6CP_IFTOKEN:        /* Interface-Token */
		dprintf(D_DT, "	{IFTOKEN: len=%d}\n", opt->po_len);
		if (opt->po_len != 2 + 2 * sizeof(long)) {
			dprintf(D_DT, "	{ADDRS: len=%d (should be %d)}\n",
			    opt->po_len,2 + 2 * sizeof(long));
			return (OPT_REJ);
		}
		dprintf(D_DT, "	{IFTOKEN: len=%d, dst=%04x, "
		    "src=%04x}\n",
		    opt->po_len,
		    ntohl((&opt->po_data.l)[0]), ntohl((&opt->po_data.l)[1]));

		res = OPT_OK;

		/* Handle the destination's address */

		if (ppp_ipv6cp_getaddr(&a, ppp, 1) == 0) {
			/* Our dst addr undefined -- accept peer's */
			dprintf(D_DT, "	*SET DST TOKEN <to %04x>*\n",
			    ntohl(*(u_long *)&opt->po_data.c[0]));
			memset(&a, 0, sizeof(a));
			a.sin6_family = AF_INET6;
			a.sin6_len = sizeof(struct sockaddr_in6);
			a.sin6_addr.s6_addr16[0] = htons(0xfe80);
			a.sin6_addr.s6_addr32[3] = ntohl((&opt->po_data.l)[0]);
			ppp_ipv6cp_setaddr(ppp, &a, 1);
		} else if (opt->po_data.l != a.sin6_addr.s6_addr32[3]) {
			dprintf(D_DT, "	*CHG DST ADDR <from %04x\n",
				ntohl((&opt->po_data.l)[0]));
			opt->po_data.l = a.sin6_addr.s6_addr32[3];
			dprintf(D_DT, "	to %04x>\n",
				ntohl((&opt->po_data.l)[0]));
			res = OPT_NAK;
		}

		/* Handle what the destination thinks our address is */

		b.sin6_family = AF_INET6;
		b.sin6_len = sizeof(struct sockaddr_in6);
		b.sin6_addr.s6_addr16[0] = htons(0xfe80);
		b.sin6_addr.s6_addr32[3] = ntohl((&opt->po_data.l)[1]);
		if (((ppp->ppp_flags & PPP_CHGADDR) && (&opt->po_data.l)[1])
		 || ppp_ipv6cp_getaddr(&a, ppp, 0) == 0) {
			/* Our src addr undefined -- accept peer's */
			dprintf(D_DT, "	*SET SRC ADDR <to %04x>*\n",
				ntohl((&opt->po_data.l)[1]));
			ppp_ipv6cp_setaddr(ppp, &b, 0);
		} else if (b.sin6_addr.s6_addr32[3]
			!= a.sin6_addr.s6_addr32[3]) {
			dprintf(D_DT, "	*WRONG SRC ADDR <from %04x",
				ntohl((&opt->po_data.l)[1]));
			(&opt->po_data.l)[1] = a.sin6_addr.s6_addr32[3];
			dprintf(D_DT, " to %04x>\n",
				ntohl((&opt->po_data.l)[1]));
			res = OPT_NAK;
		}

		return (res);

#if 0
	case IPV6CP_CPROT:        /* IP-Compression */
		dprintf(D_DT, "	{CPROT: len=%d prot=0x%x ms=%d c=%d}\n",
		    opt->po_len, ntohs(opt->po_data.s),
		    opt->po_data.c[2], opt->po_data.c[3]);
		if (opt->po_len != 4 && opt->po_len != 6)
			return (OPT_REJ);
		if (!(ppp->ppp_flags & PPP_TCPC))
			return (OPT_REJ);
		res = OPT_OK;
		if (ntohs(opt->po_data.s) != PPP_VJC) {
			opt->po_data.s = htons(PPP_VJC);
			res = OPT_NAK;
		}
		if (opt->po_len == 6) {         /* rfc1332? */
			if (opt->po_data.c[2] >= MAX_STATES) {
				opt->po_data.c[2] = MAX_STATES - 1;
				res = OPT_NAK;
			}
			if (opt->po_data.c[3] != 0) {
				opt->po_data.c[3] = 0;
				res = OPT_NAK;
			}
		}
		if (res == OPT_OK)
			ppp->ppp_txflags |= PPP_TCPC;
		return (res);
#endif
	}
	/*NOTREACHED*/
}

/*
 * Get IP address from the address list.
 * Returns 0 if there is no address.
 */
int
ppp_ipv6cp_getaddr(addr, ppp, dst)
	struct sockaddr_in6 *addr;
	ppp_t *ppp;
	int dst;
{
        struct ifreq ifr;
        int r;
        struct sockaddr_in6 *sin6;

	memset(&ifr, 0, sizeof(ifr));

	strncpy(ifr.ifr_name, (ppp->ppp_pifname[0] == '\0') ?
	    ppp->ppp_ifname : ppp->ppp_pifname, sizeof (ifr.ifr_name));
	sin6 = (struct sockaddr_in6 *) &ifr.ifr_addr;

	r = ioctl(skt, dst ? SIOCGIFDSTADDR_IN6 : SIOCGIFADDR_IN6,
		(caddr_t) &ifr);
	if (0 <= r && sin6->sin6_family == AF_INET6
	 && sin6->sin6_len != sizeof(struct sockaddr_in6)) {
		/* ioctl success */
		memcpy(addr, sin6, sin6->sin6_len);
		return 1;
	}

	/* generate interface token */
	memset(addr, 0, sizeof(struct sockaddr_in6));
	addr->sin6_family = AF_INET6;
	addr->sin6_len = sizeof(struct sockaddr_in6);
	addr->sin6_addr.s6_addr16[0] = htonl(0xfe80);
	addr->sin6_addr.s6_addr32[3] = dst ? ppp->ppp_them6 : ppp->ppp_us6;
	return 1;
}

/*
 * Set remote interface address
 */
static void
ppp_ipv6cp_setaddr(ppp, a, dst)
	ppp_t *ppp;
	struct sockaddr_in6 *a;
	int dst;
{
        struct in6_ifreq ifr;
        struct sockaddr_in6 *sin6;
	char *ifname;

	ifname = ppp->ppp_pifname[0] ? ppp->ppp_pifname : ppp->ppp_ifname;

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, ifname, sizeof (ifr.ifr_name));
	sin6 = (struct sockaddr_in6 *) &ifr.ifr_addr;
	memcpy(sin6, a, a->sin6_len);

	if (ioctl(skt, dst ? SIOCSIFDSTADDR_IN6 : SIOCSIFADDR_IN6,
			(caddr_t) &ifr) < 0) {
		warn(dst ? "%s: setting dst" : "%s: setting src", ifname);
	}
}
